import json
from typing import Dict, Text, Any

from sqlalchemy import Column, String, ForeignKey, Integer, Float, Boolean
from sqlalchemy.orm import relationship

from rasax.community.database.base import Base


class Conversation(Base):
    """Stores the user's conversation and its metadata."""

    __tablename__ = "conversation"

    sender_id = Column(String, primary_key=True)
    number_user_messages = Column(Integer, default=0)
    latest_input_channel = Column(String)
    latest_event_time = Column(Float)  # latest event time as unix timestamp
    in_training_data = Column(Boolean, default=True)
    minimum_action_confidence = Column(Float, default=1.0)
    evaluation = Column(String)
    interactive = Column(Boolean, default=False)

    events = relationship(
        "ConversationEvent", cascade="all, delete-orphan", back_populates="conversation"
    )

    unique_policies = relationship(
        "ConversationPolicyMetadata",
        cascade="all, delete-orphan",
        back_populates="conversation",
    )
    unique_actions = relationship(
        "ConversationActionMetadata",
        cascade="all, delete-orphan",
        back_populates="conversation",
    )
    unique_intents = relationship(
        "ConversationIntentMetadata",
        cascade="all, delete-orphan",
        back_populates="conversation",
    )
    unique_entities = relationship(
        "ConversationEntityMetadata",
        cascade="all, delete-orphan",
        back_populates="conversation",
    )

    corrected_messages = relationship(
        "ConversationMessageCorrection",
        cascade="all, delete-orphan",
        back_populates="conversation",
    )

    def as_dict(self) -> Dict[Text, Any]:
        return {
            "sender_id": self.sender_id,
            "latest_event_time": self.latest_event_time,
            "latest_input_channel": self.latest_input_channel,
            "intents": [i.intent for i in self.unique_intents],
            "actions": [a.action for a in self.unique_actions],
            "minimum_action_confidence": self.minimum_action_confidence,
            "in_training_data": self.in_training_data,
            "policies": [p.policy for p in self.unique_policies],
            "n_user_messages": self.number_user_messages,
            "flagged_messages": [e.timestamp for e in self.events if e.is_flagged],
            "unflagged_messages": [e.timestamp for e in self.events if e.is_unflagged],
            "corrected_messages": [
                {"message_timestamp": c.message_timestamp, "intent": c.intent}
                for c in self.corrected_messages
            ],
            "interactive": self.interactive,
        }


class ConversationEvent(Base):
    """Stores a single event which happened during a conversation."""

    __tablename__ = "conversation_event"

    id = Column(Integer, primary_key=True)
    conversation_id = Column(String, ForeignKey("conversation.sender_id"))
    conversation = relationship("Conversation", back_populates="events")

    type_name = Column(String, nullable=False)
    timestamp = Column(Float)  # time of the event as unix timestamp
    intent_name = Column(String)
    action_name = Column(String)
    policy = Column(String)
    is_flagged = Column(Boolean, default=False)
    is_unflagged = Column(Boolean, default=False)
    data = Column(String)
    message_log = relationship("MessageLog", back_populates="event", uselist=False)
    evaluation = Column(String)


class MessageLog(Base):
    """Stores the intent classification results of the user messages."""

    __tablename__ = "message_log"

    id = Column(Integer, primary_key=True)
    hash = Column(String)
    model_id = Column(Integer, ForeignKey("model.id"))
    model = relationship("Model", back_populates="message_logs")
    archived = Column(Boolean, default=False)
    time = Column(Float)  # time of the log as unix timestamp
    text = Column(String)
    intent = Column(String)
    confidence = Column(Float)
    intent_ranking = Column(String)
    entities = Column(String)

    event_id = Column(Integer, ForeignKey("conversation_event.id"))
    event = relationship(
        "ConversationEvent", uselist=False, back_populates="message_log"
    )

    def as_dict(self) -> Dict[Text, Any]:
        return {
            "id": self.id,
            "time": self.time,
            "model": self.model.name if self.model else None,
            "hash": self.hash,
            "project_id": self.model.project_id if self.model else None,
            "team": self.model.project.team if self.model else None,
            "user_input": {
                "text": self.text,
                "intent": {"name": self.intent, "confidence": self.confidence},
                "intent_ranking": json.loads(self.intent_ranking),
                "entities": json.loads(self.entities),
            },
        }


class ConversationPolicyMetadata(Base):
    """Stores the distinct set of used policies in a conversation."""

    __tablename__ = "conversation_policy_metadata"

    conversation_id = Column(
        String, ForeignKey("conversation.sender_id"), primary_key=True
    )
    policy = Column(String, primary_key=True)
    conversation = relationship("Conversation", back_populates="unique_policies")


class ConversationActionMetadata(Base):
    """Stores the distinct set of used actions in a conversation."""

    __tablename__ = "conversation_action_metadata"

    conversation_id = Column(
        String, ForeignKey("conversation.sender_id"), primary_key=True
    )
    action = Column(String, primary_key=True)
    conversation = relationship("Conversation", back_populates="unique_actions")


class ConversationIntentMetadata(Base):
    """Stores the distinct set of used intents in a conversation."""

    __tablename__ = "conversation_intent_metadata"

    conversation_id = Column(
        String, ForeignKey("conversation.sender_id"), primary_key=True
    )

    intent = Column(String, primary_key=True)
    conversation = relationship("Conversation", back_populates="unique_intents")


class ConversationEntityMetadata(Base):
    """Stores the distinct set of used entities in a conversation."""

    __tablename__ = "conversation_entity_metadata"

    conversation_id = Column(
        String, ForeignKey("conversation.sender_id"), primary_key=True
    )

    entity = Column(String, primary_key=True)
    conversation = relationship("Conversation", back_populates="unique_entities")


class ConversationMessageCorrection(Base):
    """Stores post hoc corrections of intents in a conversation."""

    __tablename__ = "conversation_message_correction"

    conversation_id = Column(
        String, ForeignKey("conversation.sender_id"), primary_key=True
    )

    # time of the message correction as unix timestamp
    message_timestamp = Column(Float, primary_key=True)
    intent = Column(String)
    conversation = relationship("Conversation", back_populates="corrected_messages")
